import * as React from 'react'
import {render} from '@testing-library/react'
import Downshift from '../'

beforeEach(() => jest.spyOn(console, 'error').mockImplementation(() => {}))
afterEach(() => console.error.mockRestore())

const Menu = ({innerRef, ...rest}) => <div ref={innerRef} {...rest} />
const RefMenu = React.forwardRef((props, ref) => <div ref={ref} {...props} />)

test('using a composite component and calling getMenuProps without a refKey results in an error', () => {
  const MyComponent = () => (
    <Downshift
      children={({getMenuProps}) => (
        <div>
          <Menu {...getMenuProps()} />
        </div>
      )}
    />
  )
  render(<MyComponent />)
  expect(console.error.mock.calls[1][0]).toMatchSnapshot()
})

test('not applying the ref prop results in an error', () => {
  const MyComponent = () => (
    <Downshift
      children={({getMenuProps}) => {
        getMenuProps()
        return (
          <div>
            <ul />
          </div>
        )
      }}
    />
  )
  render(<MyComponent />)
  expect(console.error.mock.calls[0][0]).toMatchSnapshot()
})

test('renders fine when rendering a composite component and applying getMenuProps properly', () => {
  const MyComponent = () => (
    <Downshift
      children={({getMenuProps}) => (
        <div>
          <Menu {...getMenuProps({refKey: 'innerRef'})} />
        </div>
      )}
    />
  )
  render(<MyComponent />)
  expect(console.error.mock.calls).toHaveLength(0)
})

test('using a composite component and calling getMenuProps without a refKey does not result in an error if suppressRefError is true', () => {
  const MyComponent = () => (
    <Downshift
      children={({getMenuProps}) => (
        <div>
          <Menu {...getMenuProps({}, {suppressRefError: true})} />
        </div>
      )}
    />
  )
  render(<MyComponent />)
  expect(console.error.mock.calls).toHaveLength(0)
})

test('returning a DOM element and calling getMenuProps with a refKey does not result in an error if suppressRefError is true', () => {
  const MyComponent = () => (
    <Downshift
      children={({getMenuProps}) => (
        <div>
          <ul {...getMenuProps({refKey: 'blah'}, {suppressRefError: true})} />
        </div>
      )}
    />
  )
  render(<MyComponent />)
  expect(console.error.mock.calls).toHaveLength(1)
})

test('not applying the ref prop results in an error does not result in an error if suppressRefError is true', () => {
  const MyComponent = () => (
    <Downshift
      children={({getMenuProps}) => {
        getMenuProps({}, {suppressRefError: true})
        return (
          <div>
            <ul />
          </div>
        )
      }}
    />
  )
  render(<MyComponent />)
  expect(console.error.mock.calls).toHaveLength(0)
})

test('renders fine when rendering a composite component and applying getMenuProps properly even if suppressRefError is true', () => {
  const MyComponent = () => (
    <Downshift
      children={({getMenuProps}) => (
        <div>
          <Menu
            {...getMenuProps({refKey: 'innerRef'}, {suppressRefError: true})}
          />
        </div>
      )}
    />
  )
  render(<MyComponent />)
  expect(console.error.mock.calls).toHaveLength(0)
})

test('has access to element when a ref function is passed to getMenuProps', () => {
  const ref = {current: null}

  const MyComponent = () => {
    return (
      <Downshift
        children={({getMenuProps}) => (
          <div>
            <RefMenu
              {...getMenuProps({
                ref: e => {
                  ref.current = e
                },
              })}
            />
          </div>
        )}
      />
    )
  }

  render(<MyComponent />)
  expect(ref.current).not.toBeNull()
  expect(ref.current).toBeInstanceOf(HTMLDivElement)
})
